/*
 * drivers/pcmcia/trex.c PowerBook 190 1900 5300 PCMCIA socket driver
 *
 * Copyright (C) 2003 Tony Lindgren <tony@atomide.com>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * NOTE: 
 * 
 * Please make sure you have patched the following files:
 * 1. You must edit the headers in Linux kernel and pcmcia-cs 
 *    until a proper fix is found. Make sure ioaddr_t in cs_types.h
 *    is u_int in both places.
 * 2. Patch drivers/pcmcia/cs.c to reset the memory window type to
 *    IO memory instead of attribute memory after attribute window 
 *    release.
 * 3. Undefine CONFIG_NBPMAC in drivers/net/8390.h if using cards
 *    that use 8390 module, such as pcnet_cs. This means that
 *    the nubus 8390 driver does not currently work same time with
 *    pcnet_cs.
 */

#define MODULE

#include <linux/init.h>
#include <linux/sched.h>
#include <linux/interrupt.h>
#include <linux/delay.h>
#include <linux/module.h>
#include <linux/ioport.h>
#include <linux/mm.h>
#include <linux/vmalloc.h>

#include <linux/tqueue.h>
#include <linux/timer.h>
#include <linux/notifier.h>

#include <pcmcia/version.h>
#include <pcmcia/cs_types.h>
#include <pcmcia/ss.h>
#include <pcmcia/cs.h>
#include <pcmcia/bus_ops.h>

#include <asm/io.h>

#include "trex.h"

#define PCMCIA_DEBUG
#ifdef PCMCIA_DEBUG
#define DEBUG(args...)	printk(KERN_DEBUG args)
#else
#define DEBUG(args...)
#endif

static struct tq_struct trex_pcmcia_task;

static struct trex_socket_cfg {
	volatile u_char * cfg;

	volatile u_char * cunknown0;
	volatile u_char * cunknown1;
	volatile u_char * cunknown2;
	volatile u_char * cunknown3;
	volatile u_char * cunknown4;
	volatile u_char * cunknown5;
	volatile u_char * cunknown6;
	volatile u_char * cunknown7;
	volatile u_char * cunknown8;
	volatile u_char * cunknown9;
	volatile u_char * cunknown10;
	volatile u_char * cunused11;
	volatile u_char * cunused12;
	volatile u_char * cunused13;
	volatile u_char * cunused14;
	volatile u_char * cunused15;

	int ints;
} trex_socket_cfg;

static struct trex_socket_cfg * trex = &trex_socket_cfg;

static struct trex_skt {
	int number;
	
	u_int physbase;

	volatile u_char * unknown0;
	volatile u_char * socket_config;	/* rw */
	volatile u_char * unknown2;	/* rw */
	volatile u_char * card_ready;
	volatile u_char * unknown4;	/* rw */
	volatile u_char * card_status;
	volatile u_char * unknown6;	/* rw */
	volatile u_char * unknown7;
	volatile u_char * unknown8;	/* rw */
	volatile u_char * int_status;	/* rw */
	volatile u_char * unknown10;	/* rw */
	volatile u_char * unknown11;
	volatile u_char * unknown12;
	volatile u_char * unknown13;
	volatile u_char * unknown14;
	volatile u_char * int_disable;	/* rw */

	void * common_mem;	/* Common memory (or attribute memory) */
	void * io_mem;      	/* Separate IO window */

	u_int state;

	void (*handler)(void *, u_int);
	void *handler_info;

	u_int int_pending;
	u_int ev_pending;
	spinlock_t ev_lock;
};

static struct trex_skt trex_sockets[TREX_NUM_SOCKETS];

static int enable_card(int sock);
static int __trex_get_status(struct trex_skt *skt);

static void sock_dump_common(void)
{
	DEBUG("trex: common cfg dump:\n");
	DEBUG("  00 01 02 03 04 05 06 07 08 09 10 11 12 13 14 15\n");
	DEBUG("  %02x %02x %02x %02x %02x %02x %02x %02x "
	       "%02x %02x %02x %02x %02x %02x %02x %02x\n",
	       inb(trex->cunknown0),
	       inb(trex->cunknown1),
	       inb(trex->cunknown2),
	       inb(trex->cunknown3),
	       inb(trex->cunknown4),
	       inb(trex->cunknown5),
	       inb(trex->cunknown6),
	       inb(trex->cunknown7),
	       inb(trex->cunknown8),
	       inb(trex->cunknown9),
	       inb(trex->cunknown10),
	       inb(trex->cunused11),
	       inb(trex->cunused12),
	       inb(trex->cunused13),
	       inb(trex->cunused14),
	       inb(trex->cunused15));
}

static void sock_dump_cfg(int sock)
{
	DEBUG("trex: socket %i cfg dump:\n", sock);
	DEBUG("  00 01 02 03 04 05 06 07 08 09 10 11 12 13 14 15\n");
	DEBUG("  %02x %02x %02x %02x %02x %02x %02x %02x "
	       "%02x %02x %02x %02x %02x %02x %02x %02x\n",
	       inb(trex_sockets[sock].unknown0),
	       inb(trex_sockets[sock].socket_config),
	       inb(trex_sockets[sock].unknown2),
	       inb(trex_sockets[sock].card_ready),
	       inb(trex_sockets[sock].unknown4),
	       inb(trex_sockets[sock].card_status),
	       inb(trex_sockets[sock].unknown6),
	       inb(trex_sockets[sock].unknown7),
	       inb(trex_sockets[sock].unknown8),
	       inb(trex_sockets[sock].int_status),
	       inb(trex_sockets[sock].unknown10),
	       inb(trex_sockets[sock].unknown11),
	       inb(trex_sockets[sock].unknown12),
	       inb(trex_sockets[sock].unknown13),
	       inb(trex_sockets[sock].unknown14),
	       inb(trex_sockets[sock].int_disable));
}

static inline int trex_card_inserted(int status)
{
	if ((status & CARD_STATUS_INSERTED_BIT2) ||
	    (status & CARD_STATUS_INSERTED_BIT3))
		return 0;
	else
		return 1;
}

static inline int trex_card_ready(int status)
{
	if ((status & CARD_STATUS_READY_BIT0) &&
	    (status & CARD_STATUS_READY_BIT1))
		return 1;
	else
		return 0;
}

static void trex_pcmcia_task_handler(void *data)
{
	int i;

	for (i = 0; i < TREX_NUM_SOCKETS; i++) {
		u_long flags;
		u_int events = 0;
		u_char status = 0;
		struct trex_skt *skt = &trex_sockets[i];

		if (!skt)
			continue;

		spin_lock_irqsave(&skt->ev_lock, flags);
		status = skt->int_pending;
		skt->int_pending = 0;
		events = skt->ev_pending;
		skt->ev_pending = 0;
		spin_unlock_irqrestore(&skt->ev_lock, flags);

		if (skt->handler && events) {			
			skt->handler(skt->handler_info, events);
		}
	}
}

static struct tq_struct trex_pcmcia_task = {
	routine: trex_pcmcia_task_handler
};

static void trex_interrupt(u_int irq)
{
	int i;
	u_int events;

	for (i = 0; i < TREX_NUM_SOCKETS; i++) {
		u_char status;
		struct trex_skt *skt = &trex_sockets[i];

		status = inb(skt->int_status);
		outb(0, skt->int_status);
		outb(0, skt->unknown7);
		//printk("INT: 0x%02x\n", status);

#if 0
		if (i == 1) {
			printk("%02x %02x %02x %02x %02x %02x %02x %02x "
			       "%02x %02x %02x %02x %02x %02x %02x %02x\n",
			       inb(skt->unknown0), inb(skt->socket_config),
			       inb(skt->unknown2), inb(skt->card_ready),
			       inb(skt->unknown4), inb(skt->card_status),
			       inb(skt->unknown6), inb(skt->unknown7),
			       inb(skt->unknown8), inb(skt->int_status),
			       inb(skt->unknown10), inb(skt->unknown11),
			       inb(skt->unknown12), inb(skt->unknown13),
			       inb(skt->unknown14), inb(skt->int_disable));
		}
#endif
		if (status == 0 || status == 0x10)
			continue;

		events = __trex_get_status(skt);

		/* Check for insertion or ejection */
		if (status & (1 << 2)) {
			if (events == 0) {
				events |= SS_DETECT | SS_EJECTION;
			} else
				events |= SS_INSERTION;
		}

		if (events == 0) {
			continue;
		}

		spin_lock(&skt->ev_lock);
		skt->int_pending |= status;
		skt->ev_pending |= events;
		spin_unlock(&skt->ev_lock);

		schedule_task(&trex_pcmcia_task);
	}
}

static int trex_sock_init(u_int sock)
{
	return 0;
}

static int trex_suspend(u_int sock)
{
	return 0;
}

static int trex_register_callback(u_int sock, void (*handler)(void *, u_int),
				  void *info)
{
	struct trex_skt *skt = &trex_sockets[sock];

	skt->handler_info = info;
	skt->handler = handler;

	return 0;
}

static inline void trex_use_io_mem(struct trex_skt *skt)
{
	outb(inb(skt->socket_config) & ~U1_ATTR_MEM, skt->socket_config);
}

static inline void trex_use_attr_mem(struct trex_skt *skt)
{
	outb(inb(skt->socket_config) | U1_ATTR_MEM, skt->socket_config);
}

static int trex_inquire_socket(u_int sock, socket_cap_t *cap)
{
	struct trex_skt *skt = &trex_sockets[sock];

	cap->features = (SS_CAP_PAGE_REGS | SS_CAP_PCCARD
			 | SS_CAP_STATIC_MAP | SS_CAP_MEM_ALIGN);

	cap->irq_mask = 0;
	cap->map_size = PAGE_SIZE;
	cap->pci_irq = TREX_IRQ;
	cap->cb_dev = NULL;
	cap->bus = NULL;
	cap->io_offset = (unsigned long)skt->io_mem;

	return 0;
}

static int __trex_get_status(struct trex_skt *skt)
{
	int card_status;
	u_int trex_state = 0;

	card_status = (int)inb(skt->card_status);

	if (trex_card_inserted(card_status)) {
		sock_dump_cfg(skt->number);
		enable_card(skt->number);
		trex_state |= SS_DETECT;
	}

	if ((trex_state & SS_DETECT) && 
	    trex_card_ready(card_status)) {
		trex_state |= SS_READY;
		printk("XXX enabling card interrupts\n");
		outb(0xff, skt->unknown7);
		outb(0xff, skt->card_ready);
		outb(0xff, skt->unknown4);
		outb(0xff, skt->card_status);
	}

	trex_state &= ~SS_BATWARN;
	trex_state &= ~SS_BATDEAD;
	trex_state &= ~SS_XVCARD;
	trex_state &= ~SS_PENDING;

        DEBUG("Trex: Status: skt%d: status: %08x -> %s %s %s %s %s %s)\n",
	      skt->number, card_status,
	      trex_state & SS_READY   ? "rdy" : "---",
	      trex_state & SS_DETECT  ? "det" : "---",
	      trex_state & SS_BATWARN ? "bw"  : "--",
	      trex_state & SS_BATDEAD ? "bd"  : "--",
	      trex_state & SS_3VCARD  ? "3v"  : "--",
	      trex_state & SS_XVCARD  ? "xv"  : "--");

	return trex_state;
}

static int trex_get_status(u_int sock, u_int *valp)
{
	struct trex_skt *skt = &trex_sockets[sock];
	*valp = __trex_get_status(skt);
	return 0; /* not used */
}

static int trex_get_socket(u_int sock, socket_state_t *state)
{
	return -EINVAL;
}

static int trex_set_socket(u_int sock, socket_state_t *state)
{
	return 0;
}

static int trex_get_io_map(u_int sock, struct pccard_io_map *io)
{
	return -EINVAL;
}

static int trex_set_io_map(u_int sock, struct pccard_io_map *map)
{
	u_long start;
	struct trex_skt *skt = &trex_sockets[sock];

	start = map->start;

	if (map->stop == 1)
		map->stop = PAGE_SIZE - 1;

	map->start = (u_long)skt->io_mem;
	map->stop = map->start + (map->stop - start);

	/* 
	 * NOTE: If typedef for ioaddr_t in cs_types.h is u_char, we will oops!
	 *       Must edit in Linux headers AND pcmcia-cs headers.
	 */ 
	printk("trex: io_mem: 0x%08x, map->start: 0x%08x\n", 
	       (u_int)skt->io_mem, map->start);

	return 0;
}

static int trex_get_mem_map(u_int sock, struct pccard_mem_map *mem)
{
	return -EINVAL;
}

static int trex_set_mem_map(u_int sock, struct pccard_mem_map *mem)
{
	struct trex_skt *skt = &trex_sockets[sock];
	u_long start;

	start = mem->sys_start;
	if (mem->sys_stop == 0)
		mem->sys_stop = PAGE_SIZE - 1;

	if (mem->flags & MAP_ATTRIB) {
		trex_use_attr_mem(skt);
		mem->sys_start = skt->physbase + mem->card_start;
	} else {
		trex_use_io_mem(skt);
		mem->sys_start = skt->physbase + mem->card_start;
	}

	return 0;
}

static void trex_proc_setup(u_int sock, struct proc_dir_entry *base)
{
	/* nothing */
}

static struct pccard_operations trex_operations = {
	init:			trex_sock_init,
	suspend:		trex_suspend,
	register_callback:	trex_register_callback,
	inquire_socket:		trex_inquire_socket,
	get_status:		trex_get_status,
	get_socket:		trex_get_socket,
	set_socket:		trex_set_socket,
	get_io_map:		trex_get_io_map,
	set_io_map:		trex_set_io_map,
	get_mem_map:		trex_get_mem_map,
	set_mem_map:		trex_set_mem_map,
	proc_setup:		trex_proc_setup
};

static int trex_init_sock(int sock)
{
	volatile u_char * cfg_base;
	u_int mem_base;

	trex_sockets[sock].number = sock;
	
	cfg_base = trex->cfg;

	if (sock)
		cfg_base += 16;

	trex_sockets[sock].unknown0 = (volatile u_char *)cfg_base + 0x0;
	trex_sockets[sock].socket_config = (volatile u_char *)cfg_base + 0x1;
	trex_sockets[sock].unknown2 = (volatile u_char *)cfg_base + 0x2;
	trex_sockets[sock].card_ready = (volatile u_char *)cfg_base + 0x3;
	trex_sockets[sock].unknown4 = (volatile u_char *)cfg_base + 0x4;
	trex_sockets[sock].card_status = (volatile u_char *)cfg_base + 0x5;
	trex_sockets[sock].unknown6 = (volatile u_char *)cfg_base + 0x6;
	trex_sockets[sock].unknown7 = (volatile u_char *)cfg_base + 0x7;
	trex_sockets[sock].unknown8 = (volatile u_char *)cfg_base + 0x8;
	trex_sockets[sock].int_status = (volatile u_char *)cfg_base + 0x9;
	trex_sockets[sock].unknown10 = (volatile u_char *)cfg_base + 0xa;
	trex_sockets[sock].unknown11 = (volatile u_char *)cfg_base + 0xb;
	trex_sockets[sock].unknown12 = (volatile u_char *)cfg_base + 0xc;
	trex_sockets[sock].unknown13 = (volatile u_char *)cfg_base + 0xd;
	trex_sockets[sock].unknown14 = (volatile u_char *)cfg_base + 0xe;
	trex_sockets[sock].int_disable = (volatile u_char *)cfg_base + 0xf;

	mem_base = TREX_SLOT0_BASE;
	if (sock)
		mem_base += 0x8000000;

	trex_sockets[sock].physbase = mem_base;

	trex_sockets[sock].io_mem = (void *)ioremap(mem_base + 0x4000000,
						     0x4000000);
	if (!trex_sockets[sock].io_mem) {
		printk("trex: could not ioremap io_mem\n");
		goto out_nomem;
	}

	trex_sockets[sock].common_mem = (void *)ioremap(mem_base + 0x0000000, 
						      0x4000000);
	if (!trex_sockets[sock].common_mem) {
		printk("trex: could not ioremap common_mem\n");
		goto out_nomem;
	}

	printk("trex: Sock %i io_mem: 0x%08x (0x%08x) "
	       "common_mem: 0x%08x (0x%08x)\n",
	       sock, (u_int)trex_sockets[sock].io_mem, mem_base + 0x4000000,
	       (u_int)trex_sockets[sock].common_mem, mem_base + 0x0000000);

	trex_sockets[sock].state = 0;

	return 0;

 out_nomem:
	if (trex_sockets[sock].io_mem)
		iounmap((void *)trex_sockets[sock].io_mem);

	if (trex_sockets[sock].common_mem)
		iounmap((void *)trex_sockets[sock].common_mem);

	return -ENOMEM;
}

static int enable_card(int sock)
{
	struct trex_skt *skt = &trex_sockets[sock];

	outb(U1_ATTR_MASK, skt->socket_config);
	udelay(1000);			/* Wait needed for modem */
	outb(U2_MASK, skt->unknown2);

	return 0;
}

static int disable_card(int sock)
{
	struct trex_skt *skt = &trex_sockets[sock];

	outb(0, skt->socket_config);
	outb(0, skt->unknown2);
	//outb(0, skt->int_disable);	/* Writing 0 disables ints */

	return 0;
}

static int __init trex_init(void)
{
	int nr, ret;
	volatile u_char * cfg_base;

	printk("Initializing TRex PCMCIA controller\n");

	trex->ints = 0;

	trex->cfg = (volatile u_char *)ioremap(TREX_CFG, 0x100);
	if (!trex->cfg) {
		printk("trex: could not ioremap configration registers\n");
		return -ENOMEM;
	}

	cfg_base = trex->cfg + 32;

	/* Common configuration area */
	trex->cunknown0 = (volatile u_char *)cfg_base + 0x0;
	trex->cunknown1 = (volatile u_char *)cfg_base + 0x1;
	trex->cunknown2 = (volatile u_char *)cfg_base + 0x2;
	trex->cunknown3 = (volatile u_char *)cfg_base + 0x3;
	trex->cunknown4 = (volatile u_char *)cfg_base + 0x4;
	trex->cunknown5 = (volatile u_char *)cfg_base + 0x5;
	trex->cunknown6 = (volatile u_char *)cfg_base + 0x6;
	trex->cunknown7 = (volatile u_char *)cfg_base + 0x7;
	trex->cunknown8 = (volatile u_char *)cfg_base + 0x8;
	trex->cunknown9 = (volatile u_char *)cfg_base + 0x9;
	trex->cunknown10 = (volatile u_char *)cfg_base + 0xa;
	trex->cunused11 = (volatile u_char *)cfg_base + 0xb;
	trex->cunused12 = (volatile u_char *)cfg_base + 0xc;
	trex->cunused13 = (volatile u_char *)cfg_base + 0xd;
	trex->cunused14 = (volatile u_char *)cfg_base + 0xe;
	trex->cunused15 = (volatile u_char *)cfg_base + 0xf;

	/* Socket configurations */
	for (nr = 0; nr < TREX_NUM_SOCKETS; nr++) {
		ret = trex_init_sock(nr);
	}

	if (ret)
		return ret;

#if 0
	printk("Values on load:\n");
	sock_dump_common();
	sock_dump_cfg(0);
	sock_dump_cfg(1);
#endif

	enable_card(0);
	enable_card(1);

	ret = request_irq(TREX_IRQ, (void *)trex_interrupt, 0, "trex", NULL);
	if (ret) {
		printk(KERN_ERR "trex: request_irq failed %d\n", ret);
		return 1;
	}

	ret = register_ss_entry(nr, &trex_operations);

	return 0;
}

static void __exit trex_exit(void)
{
	int i;

	printk("Unloading TRex PCMCIA controller\n");

	for (i = 0; i < TREX_NUM_SOCKETS; i++) {
		disable_card(i);
	}

	unregister_ss_entry(&trex_operations);

	free_irq(TREX_IRQ, NULL);

	for (i = 0; i < TREX_NUM_SOCKETS; i++) {
		struct trex_skt *skt = &trex_sockets[i];

		if (skt->common_mem)
			iounmap((void *)skt->common_mem);

		if (skt->io_mem)
			iounmap((void *)skt->io_mem);
	}

	if (trex->cfg)
		iounmap((void *)trex->cfg);
	
}

MODULE_LICENSE("GPL");
module_init(trex_init);
module_exit(trex_exit);
